# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../../nbs/llama_index.indices.knowledge_graph.base.ipynb.

# %% auto 0
__all__ = ['logger', 'KnowledgeGraphIndex', 'patch_knowledge_graph_index']

# %% ../../../../nbs/llama_index.indices.knowledge_graph.base.ipynb 3
import logging
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple

from llama_index.constants import GRAPH_STORE_KEY
from llama_index.core.base_retriever import BaseRetriever
from llama_index.graph_stores.simple import SimpleGraphStore
from llama_index.graph_stores.types import GraphStore
from llama_index.indices.base import BaseIndex
from llama_index.prompts import BasePromptTemplate
from llama_index.prompts.default_prompts import DEFAULT_KG_TRIPLET_EXTRACT_PROMPT
from llama_index.schema import BaseNode, IndexNode, MetadataMode
from llama_index.service_context import ServiceContext
from llama_index.storage.docstore.types import RefDocInfo
from llama_index.storage.storage_context import StorageContext
from llama_index.utils import get_tqdm_iterable
from ...data_structs.data_structs import KG


# %% ../../../../nbs/llama_index.indices.knowledge_graph.base.ipynb 4
logger = logging.getLogger(__name__)


class KnowledgeGraphIndex(BaseIndex[KG]):
    """Knowledge Graph Index.

    Build a KG by extracting triplets, and leveraging the KG during query-time.

    Args:
        kg_triple_extract_template (BasePromptTemplate): The prompt to use for
            extracting triplets.
        max_triplets_per_chunk (int): The maximum number of triplets to extract.
        service_context (Optional[ServiceContext]): The service context to use.
        storage_context (Optional[StorageContext]): The storage context to use.
        graph_store (Optional[GraphStore]): The graph store to use.
        show_progress (bool): Whether to show tqdm progress bars. Defaults to False.
        include_embeddings (bool): Whether to include embeddings in the index.
            Defaults to False.
        max_object_length (int): The maximum length of the object in a triplet.
            Defaults to 128.
        kg_triplet_extract_fn (Optional[Callable]): The function to use for
            extracting triplets. Defaults to None.

    """

    index_struct_cls = KG

    def __init__(
        self,
        nodes: Optional[Sequence[BaseNode]] = None,
        objects: Optional[Sequence[IndexNode]] = None,
        index_struct: Optional[KG] = None,
        service_context: Optional[ServiceContext] = None,
        storage_context: Optional[StorageContext] = None,
        kg_triple_extract_template: Optional[BasePromptTemplate] = None,
        max_triplets_per_chunk: int = 10,
        include_embeddings: bool = False,
        show_progress: bool = False,
        max_object_length: int = 128,
        kg_triplet_extract_fn: Optional[Callable] = None,
        **kwargs: Any,
    ) -> None:
        """Initialize params."""
        # need to set parameters before building index in base class.
        self.include_embeddings = include_embeddings
        self.max_triplets_per_chunk = max_triplets_per_chunk
        self.kg_triple_extract_template = (
            kg_triple_extract_template or DEFAULT_KG_TRIPLET_EXTRACT_PROMPT
        )
        # NOTE: Partially format keyword extract template here.
        self.kg_triple_extract_template = (
            self.kg_triple_extract_template.partial_format(
                max_knowledge_triplets=self.max_triplets_per_chunk
            )
        )
        self._max_object_length = max_object_length
        self._kg_triplet_extract_fn = kg_triplet_extract_fn

        super().__init__(
            nodes=nodes,
            index_struct=index_struct,
            service_context=service_context,
            storage_context=storage_context,
            show_progress=show_progress,
            objects=objects,
            **kwargs,
        )

        # TODO: legacy conversion - remove in next release
        if (
            len(self.index_struct.table) > 0
            and isinstance(self.graph_store, SimpleGraphStore)
            and len(self.graph_store._data.graph_dict) == 0
        ):
            logger.warning("Upgrading previously saved KG index to new storage format.")
            self.graph_store._data.graph_dict = self.index_struct.rel_map

    @property
    def graph_store(self) -> GraphStore:
        return self._graph_store

    def as_retriever(self, **kwargs: Any) -> BaseRetriever:
        from bellek.llama_index.indices.knowledge_graph.retrievers import KGTableRetriever
        from llama_index.indices.knowledge_graph.retrievers import KGRetrieverMode
        if len(self.index_struct.embedding_dict) > 0 and "retriever_mode" not in kwargs:
            kwargs["retriever_mode"] = KGRetrieverMode.HYBRID

        return KGTableRetriever(self, object_map=self._object_map, **kwargs)

    def _extract_triplets(self, text: str) -> List[Tuple[str, str, str]]:
        if self._kg_triplet_extract_fn is not None:
            return self._kg_triplet_extract_fn(text)
        else:
            return self._llm_extract_triplets(text)

    def _llm_extract_triplets(self, text: str) -> List[Tuple[str, str, str]]:
        """Extract keywords from text."""
        response = self._service_context.llm.predict(
            self.kg_triple_extract_template,
            text=text,
        )
        return self._parse_triplet_response(
            response, max_length=self._max_object_length
        )

    @staticmethod
    def _parse_triplet_response(
        response: str, max_length: int = 128
    ) -> List[Tuple[str, str, str]]:
        knowledge_strs = response.strip().split("\n")
        results = []
        for text in knowledge_strs:
            if "(" not in text or ")" not in text or text.index(")") < text.index("("):
                # skip empty lines and non-triplets
                continue
            triplet_part = text[text.index("(") + 1 : text.index(")")]
            tokens = triplet_part.split(",")
            if len(tokens) != 3:
                continue

            if any(len(s.encode("utf-8")) > max_length for s in tokens):
                # We count byte-length instead of len() for UTF-8 chars,
                # will skip if any of the tokens are too long.
                # This is normally due to a poorly formatted triplet
                # extraction, in more serious KG building cases
                # we'll need NLP models to better extract triplets.
                continue

            subj, pred, obj = map(str.strip, tokens)
            if not subj or not pred or not obj:
                # skip partial triplets
                continue

            # Strip double quotes and Capitalize triplets for disambiguation
            subj, pred, obj = (
                entity.strip('"').capitalize() for entity in [subj, pred, obj]
            )

            results.append((subj, pred, obj))
        return results

    def _build_index_from_nodes(self, nodes: Sequence[BaseNode]) -> KG:
        """Build the index from nodes."""
        # do simple concatenation
        index_struct = self.index_struct_cls()
        nodes_with_progress = get_tqdm_iterable(
            nodes, self._show_progress, "Processing nodes"
        )
        for n in nodes_with_progress:
            triplets = self._extract_triplets(
                n.get_content(metadata_mode=MetadataMode.LLM)
            )
            logger.debug(f"> Extracted triplets: {triplets}")
            for triplet in triplets:
                subj, _, obj = triplet
                self.upsert_triplet(triplet)
                index_struct.add_node([subj, obj], n)

            if self.include_embeddings:
                triplet_texts = [str(t) for t in triplets]

                embed_model = self._service_context.embed_model
                embed_outputs = embed_model.get_text_embedding_batch(
                    triplet_texts, show_progress=self._show_progress
                )
                for rel_text, rel_embed in zip(triplet_texts, embed_outputs):
                    index_struct.add_to_embedding_dict(rel_text, rel_embed)

        return index_struct

    def _insert(self, nodes: Sequence[BaseNode], **insert_kwargs: Any) -> None:
        """Insert a document."""
        for n in nodes:
            triplets = self._extract_triplets(
                n.get_content(metadata_mode=MetadataMode.LLM)
            )
            logger.debug(f"Extracted triplets: {triplets}")
            for triplet in triplets:
                subj, _, obj = triplet
                triplet_str = str(triplet)
                self.upsert_triplet(triplet)
                self._index_struct.add_node([subj, obj], n)
                if (
                    self.include_embeddings
                    and triplet_str not in self._index_struct.embedding_dict
                ):
                    rel_embedding = (
                        self._service_context.embed_model.get_text_embedding(
                            triplet_str
                        )
                    )
                    self._index_struct.add_to_embedding_dict(triplet_str, rel_embedding)

    def upsert_triplet(self, triplet: Tuple[str, str, str]) -> None:
        """Insert triplets.

        Used for manual insertion of KG triplets (in the form
        of (subject, relationship, object)).

        Args:
            triplet (str): Knowledge triplet

        """
        self._graph_store.upsert_triplet(*triplet)

    def add_node(self, keywords: List[str], node: BaseNode) -> None:
        """Add node.

        Used for manual insertion of nodes (keyed by keywords).

        Args:
            keywords (List[str]): Keywords to index the node.
            node (Node): Node to be indexed.

        """
        self._index_struct.add_node(keywords, node)
        self._docstore.add_documents([node], allow_update=True)

    def upsert_triplet_and_node(
        self, triplet: Tuple[str, str, str], node: BaseNode
    ) -> None:
        """Upsert KG triplet and node.

        Calls both upsert_triplet and add_node.
        Behavior is idempotent; if Node already exists,
        only triplet will be added.

        Args:
            keywords (List[str]): Keywords to index the node.
            node (Node): Node to be indexed.

        """
        subj, _, obj = triplet
        self.upsert_triplet(triplet)
        self.add_node([subj, obj], node)

    def _delete_node(self, node_id: str, **delete_kwargs: Any) -> None:
        """Delete a node."""
        raise NotImplementedError("Delete is not supported for KG index yet.")

    @property
    def ref_doc_info(self) -> Dict[str, RefDocInfo]:
        """Retrieve a dict mapping of ingested documents and their nodes+metadata."""
        node_doc_ids_sets = list(self._index_struct.table.values())
        node_doc_ids = list(set().union(*node_doc_ids_sets))
        nodes = self.docstore.get_nodes(node_doc_ids)

        all_ref_doc_info = {}
        for node in nodes:
            ref_node = node.source_node
            if not ref_node:
                continue

            ref_doc_info = self.docstore.get_ref_doc_info(ref_node.node_id)
            if not ref_doc_info:
                continue

            all_ref_doc_info[ref_node.node_id] = ref_doc_info
        return all_ref_doc_info

    def get_networkx_graph(self, limit: int = 100) -> Any:
        """Get networkx representation of the graph structure.

        Args:
            limit (int): Number of starting nodes to be included in the graph.

        NOTE: This function requires networkx to be installed.
        NOTE: This is a beta feature.

        """
        try:
            import networkx as nx
        except ImportError:
            raise ImportError(
                "Please install networkx to visualize the graph: `pip install networkx`"
            )

        g = nx.Graph()
        subjs = list(self.index_struct.table.keys())

        # add edges
        rel_map = self._graph_store.get_rel_map(subjs=subjs, depth=1, limit=limit)

        added_nodes = set()
        for keyword in rel_map:
            for path in rel_map[keyword]:
                subj = keyword
                for i in range(0, len(path), 2):
                    if i + 2 >= len(path):
                        break

                    if subj not in added_nodes:
                        g.add_node(subj)
                        added_nodes.add(subj)

                    rel = path[i + 1]
                    obj = path[i + 2]

                    g.add_edge(subj, obj, label=rel, title=rel)
                    subj = obj
        return g

    @property
    def query_context(self) -> Dict[str, Any]:
        return {GRAPH_STORE_KEY: self._graph_store}


# %% ../../../../nbs/llama_index.indices.knowledge_graph.base.ipynb 5
def patch_knowledge_graph_index():
    from llama_index.data_structs.struct_type import IndexStructType
    from llama_index.indices.registry import INDEX_STRUCT_TYPE_TO_INDEX_CLASS
    INDEX_STRUCT_TYPE_TO_INDEX_CLASS[IndexStructType.KG] = KnowledgeGraphIndex



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/jerx.prompt.ipynb.

# %% auto 0
__all__ = ['log', 'DEFAULT_SYSTEM_PROMPT_TEMPLATE', 'DEFAULT_RELATION_SET_PROMPT_TEMPLATE',
           'DEFAULT_SIMPLE_SYSTEM_PROMPT_TEMPLATE', 'JERXChatFormatter', 'JERXSimpleChatFormatter']

# %% ../../nbs/jerx.prompt.ipynb 3
import random
from dataclasses import dataclass
from typing import Iterable, Callable, Any, Generator

from ..logging import get_logger

log = get_logger(__name__)

# %% ../../nbs/jerx.prompt.ipynb 6
DEFAULT_SYSTEM_PROMPT_TEMPLATE = """You are a helpful assistant that extracts up to {max_triplets} entity-relation-entity triplets from given text. Use '{delimiter}' as delimiter and provide one triplet per line. The entities in a triplet must be different.
{relation_set_prompt}
""".strip()

DEFAULT_RELATION_SET_PROMPT_TEMPLATE = """Here are the list of relations that you can use:
{relation_set}
""".strip()

@dataclass
class JERXChatFormatter:
    system_prompt_template: str = DEFAULT_SYSTEM_PROMPT_TEMPLATE
    relation_set_prompt_template: str = DEFAULT_RELATION_SET_PROMPT_TEMPLATE
    relation_set: set|None = None
    max_triplets_margin: int = 0
    delimiter: str = " | "

    def __post_init__(self):
        if self.relation_set:
            self.relation_set = sorted(self.relation_set)

    def format(self, batch: list[dict], max_triplets: int | None = None):
        if "triplets" not in batch[0]:
            assert len(batch) == 1, "Only one example is allowed when 'triplets' is not present"
        if max_triplets is None:
            if "triplets" in batch[0]:
                max_triplets = max([len(example['triplets']) for example in batch]) + self.max_triplets_margin
            else:
                max_triplets = random.randint(15, 20)
        messages = [
            self.make_system_message(max_triplets),
            *[message for example in batch for message in self.make_messages(example)],
        ]
        return {'messages': messages}

    def make_system_message(self, max_triplets: int) -> str:
        rsp = self.relation_set_prompt_template.format(relation_set=','.join(self.relation_set)) if self.relation_set else ""
        content = self.system_prompt_template.format(max_triplets=max_triplets, delimiter=self.delimiter, relation_set_prompt=rsp)
        return {"role": "system", "content": content}

    def make_messages(self, example: dict) -> Generator[dict, None, None]:
        yield {"role": "user", "content": example["text"]}
        if "triplets" in example:
            yield {"role": "assistant", "content": self._format_triplets(example["triplets"])}

    def _format_triplets(self, triplets: Iterable[str]) -> str:
        return '\n'.join(triplets)

# %% ../../nbs/jerx.prompt.ipynb 9
DEFAULT_SIMPLE_SYSTEM_PROMPT_TEMPLATE = """
You are an excellent knowledge graph construction agent. Extract knowledge triplets in the form of (subject, relation, object) from user's messages. Avoid stopwords. Use '{delimiter}' as delimiter and provide one triplet per line.
""".strip()

@dataclass
class JERXSimpleChatFormatter:
    system_prompt_template: str = DEFAULT_SIMPLE_SYSTEM_PROMPT_TEMPLATE
    delimiter: str = " | "

    def format(self, batch: list[dict]):
        if "triplets" not in batch[0]:
            assert len(batch) == 1, "Only one example is allowed when 'triplets' is not present"
        messages = [
            self.make_system_message(),
            *[message for example in batch for message in self.make_messages(example)],
        ]
        return {'messages': messages}

    def make_system_message(self) -> str:
        content = self.system_prompt_template.format(delimiter=self.delimiter)
        return {"role": "system", "content": content}

    def make_messages(self, example: dict) -> Generator[dict, None, None]:
        yield {"role": "user", "content": example["text"]}
        if "triplets" in example:
            yield {"role": "assistant", "content": self._format_triplets(example["triplets"])}

    def _format_triplets(self, triplets: Iterable[str]) -> str:
        return '\n'.join(set(triplets))



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/llama_index.data_structs.data_structs.ipynb.

# %% auto 0
__all__ = ['KG', 'patch_kg_data_struct']

# %% ../../../nbs/llama_index.data_structs.data_structs.ipynb 3
from dataclasses import dataclass, field
from typing import Dict, List, Set

from llama_index.data_structs.struct_type import IndexStructType
from llama_index.data_structs.data_structs import IndexStruct
from ...text.utils import fuzzy_match
from llama_index.schema import BaseNode

# %% ../../../nbs/llama_index.data_structs.data_structs.ipynb 4
@dataclass
class KG(IndexStruct):
    """A table of keywords mapping keywords to text chunks."""

    # Unidirectional

    # table of keywords to node ids
    table: Dict[str, Set[str]] = field(default_factory=dict)

    # TODO: legacy attribute, remove in future releases
    rel_map: Dict[str, List[List[str]]] = field(default_factory=dict)

    # TBD, should support vector store, now we just persist the embedding memory
    # maybe chainable abstractions for *_stores could be designed
    embedding_dict: Dict[str, List[float]] = field(default_factory=dict)

    # keyword match params
    keyword_match_threshold: float = 0.6

    @property
    def node_ids(self) -> Set[str]:
        """Get all node ids."""
        return set.union(*self.table.values())

    def add_to_embedding_dict(self, triplet_str: str, embedding: List[float]) -> None:
        """Add embedding to dict."""
        self.embedding_dict[triplet_str] = embedding

    def add_node(self, keywords: List[str], node: BaseNode) -> None:
        """Add text to table."""
        node_id = node.node_id
        for keyword in keywords:
            if keyword not in self.table:
                self.table[keyword] = set()
            self.table[keyword].add(node_id)

    def search_node_by_keyword(self, keyword: str) -> Dict[str, List[str]]:
        result = dict()
        for k, node_set in self.table.items():
            if fuzzy_match(keyword, k, threshold=self.keyword_match_threshold):
                result[k] = list(node_set)
        return result
    
    @classmethod
    def get_type(cls) -> IndexStructType:
        """Get type."""
        return IndexStructType.KG


# %% ../../../nbs/llama_index.data_structs.data_structs.ipynb 5
def patch_kg_data_struct():
    from llama_index.data_structs.registry import INDEX_STRUCT_TYPE_TO_INDEX_STRUCT_CLASS
     
    INDEX_STRUCT_TYPE_TO_INDEX_STRUCT_CLASS[IndexStructType.KG] =  KG



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/jerx.reward.heuristic.ipynb.

# %% auto 0
__all__ = ['log', 'compute_heuristic_reward']

# %% ../../../nbs/jerx.reward.heuristic.ipynb 3
from ...logging import get_logger

log = get_logger(__name__)

# %% ../../../nbs/jerx.reward.heuristic.ipynb 4
def compute_heuristic_reward(generation: str, delimiter: str = "|") -> float:
    lines = generation.splitlines()
    if len(lines) < 2:
        return 0

    triplets = set([line for line in lines if len(line.split(delimiter)) == 3])
    if len(triplets) > 30:
        return 0

    entities = set()
    relations = set()
    for triplet in triplets:
        subj, relation, obj = triplet.split(delimiter)
        entities.add(subj.strip())
        entities.add(obj.strip())
        relations.add(relation.strip())

    reward = 0
    if len(entities) > 5:
        reward += 0.3
    if len(relations) > 5:
        reward += 0.5

    if len(triplets) > 5:
        reward += 0.1
    elif len(triplets) >= 3:
        reward += 0.05

    if (len(triplets) / len(lines)) > 0.8:
        reward += 0.05

    return reward





# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/musique.singlehop.ipynb.

# %% auto 0
__all__ = ['make_docs', 'BaselineSingleHop', 'benchmark']

# %% ../../nbs/musique.singlehop.ipynb 3
from typing import Callable

import pandas as pd
from tqdm.auto import tqdm

from .eval import calculate_metrics, compare_answers

tqdm.pandas()

# %% ../../nbs/musique.singlehop.ipynb 4
def make_docs(example):
    ps = example["paragraphs"]
    for p in ps:
        idx = p["idx"]
        title = p["title"]
        body = p["paragraph_text"]
        is_supporting = p["is_supporting"]
        text = f"# {title}\n{body}"
        yield dict(
            text=text,
            is_supporting=is_supporting,
            parent_id=example["id"],
            idx=idx,
        )

# %% ../../nbs/musique.singlehop.ipynb 5
class BaselineSingleHop:
    def __init__(self, qa_func, retrieval_func):
        self.qa_func = qa_func
        self.retrieval_func = retrieval_func

    def _call(self, example) -> dict:
        docs = list(make_docs(example))
        question = example["question"]
        query = question
        docs = self.retrieval_func(docs, query)
        context = "\n".join(doc['text'] for doc in docs)
        result = self.qa_func(context=context, question=question)
        answer = result.get("answer")
        hop = {
            "question": question,
            "query" : query,
            "context": context,
            "answer": answer,
            "llm_output": result,
        }
        return {'answer': answer, 'hops': [hop]}

    def __call__(self, example, ignore_errors: bool = False) -> dict:
        try:
            output = self._call(example)
        except Exception as exc:
            if ignore_errors:
                id = example["id"]
                print(f"Failed to answer the question {id}\n{exc}")
                output = dict(answer="N/A", hops=[{'error': str(exc)}])
            else:
                raise
        return output

# %% ../../nbs/musique.singlehop.ipynb 6
def benchmark(
    dataf: pd.DataFrame,
    qa_func: Callable,
    retrieval_func: Callable,
    ignore_errors: bool = False,
) -> tuple[pd.DataFrame, dict]:
    pipeline = BaselineSingleHop(qa_func, retrieval_func)

    def process(example):
        output = pipeline(example, ignore_errors=ignore_errors)
        example["predicted_answer"] = output['answer']
        example["raw_output"] = output
        return example

    dataf = dataf.progress_apply(process, axis=1)
    dataf = compare_answers(dataf)
    scores = calculate_metrics(dataf)
    scores["fuzzy_match"] = dataf["fuzzy_match"].mean()
    return dataf, scores



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/musique.multihop.ipynb.

# %% auto 0
__all__ = ['make_docs', 'BaselineMultiHop', 'benchmark']

# %% ../../nbs/musique.multihop.ipynb 3
from typing import Callable

import pandas as pd
from tqdm.auto import tqdm

from .eval import calculate_metrics, compare_answers

tqdm.pandas()

# %% ../../nbs/musique.multihop.ipynb 4
def make_docs(example):
    ps = example["paragraphs"]
    for p in ps:
        idx = p["idx"]
        title = p["title"]
        body = p["paragraph_text"]
        is_supporting = p["is_supporting"]
        text = f"# {title}\n{body}"
        yield dict(
            text=text,
            is_supporting=is_supporting,
            parent_id=example["id"],
            idx=idx,
        )

# %% ../../nbs/musique.multihop.ipynb 5
class BaselineMultiHop:
    def __init__(self, qa_func, retrieval_func):
        self.qa_func = qa_func
        self.retrieval_func = retrieval_func

    def _call(self, example) -> dict:
        docs = list(make_docs(example))
        
        # First question
        question1 = example["question_decomposition"][0]["question"]
        query1 = question1
        docs1 = self.retrieval_func(docs, query1)
        context1 = "\n".join(doc['text'] for doc in docs1)
        result1 = self.qa_func(context=context1, question=question1)
        answer1 = result1.get("answer")
        hop1 = {
            "question": question1,
            "query" : query1,
            "context": context1,
            "answer": answer1,
            "llm_output": result1,
        }

        # Second question
        if answer1.strip() == "N/A":
            return {
                "answer": "N/A",
                "hops": [hop1],
            }

        question2 = example["question_decomposition"][1]["question"]
        question2 = question2.replace("#1", answer1)
        query2 = question2
        docs2 = self.retrieval_func(docs, query2)
        context2 = "\n".join(doc['text'] for doc in docs2)
        result2 = self.qa_func(context=context2, question=question2)
        answer2 = result2.get("answer")
        hop2 = {
            "question": question2,
            "query": query2,
            "context": context2,
            "answer": answer2,
            "llm_output": result2,
        }

        return {'answer': answer2, 'hops': [hop1, hop2]}

    def __call__(self, example, ignore_errors: bool = False) -> dict:
        try:
            output = self._call(example)
        except Exception as exc:
            if ignore_errors:
                id = example["id"]
                print(f"Failed to answer the question {id}\n{exc}")
                output = dict(answer="N/A", hops=[{'error': str(exc)}])
            else:
                raise
        return output

# %% ../../nbs/musique.multihop.ipynb 6
def benchmark(
    dataf: pd.DataFrame,
    qa_func: Callable,
    retrieval_func: Callable,
    ignore_errors: bool = False,
) -> tuple[pd.DataFrame, dict]:
    pipeline = BaselineMultiHop(qa_func, retrieval_func)

    def process(example):
        output = pipeline(example, ignore_errors=ignore_errors)
        example["predicted_answer"] = output['answer']
        example["raw_output"] = output
        return example

    dataf = dataf.progress_apply(process, axis=1)
    dataf = compare_answers(dataf)
    scores = calculate_metrics(dataf)
    scores["fuzzy_match"] = dataf["fuzzy_match"].mean()
    return dataf, scores



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/musique.baseline.ipynb.

# %% auto 0
__all__ = ['make_docs', 'BaselineSingleHop', 'BaselineMultiHop', 'benchmark']

# %% ../../nbs/musique.baseline.ipynb 3
import json
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor, as_completed
from typing import Callable

import pandas as pd
from tqdm.auto import tqdm

from ..jerx.reward.llm import QuestionAnsweringResult
from .eval import calculate_metrics, compare_answers

tqdm.pandas()

# %% ../../nbs/musique.baseline.ipynb 4
def make_docs(example):
    ps = example["paragraphs"]
    for p in ps:
        idx = p["idx"]
        title = p["title"]
        body = p["paragraph_text"]
        is_supporting = p["is_supporting"]
        text = f"# {title}\n{body}"
        yield dict(
            text=text,
            is_supporting=is_supporting,
            parent_id=example["id"],
            id=idx,
        )

# %% ../../nbs/musique.baseline.ipynb 5
class BaselineSingleHop:
    def __init__(self, qa_func, retrieval_func):
        self.qa_func = qa_func
        self.retrieval_func = retrieval_func

    def _call(self, example) -> QuestionAnsweringResult:
        docs = list(make_docs(example))
        retrieved_docs = self.retrieval_func(docs, example['question'])
        context = "\n\n".join([doc["text"] for doc in retrieved_docs])
        return self.qa_func(context=context, question=example['question'])

    def __call__(self, example, ignore_errors: bool = False) -> QuestionAnsweringResult:
        try:
            output = self._call(example)
        except Exception as exc:
            if ignore_errors:
                id = example['id']
                print(f"Failed to answer the question {id}\n{exc}")
                output = QuestionAnsweringResult(reasoning="", answer="N/A", raw_output=str(exc))
            else:
                raise
        return output

# %% ../../nbs/musique.baseline.ipynb 6
class BaselineMultiHop:
    def __init__(self, qa_func, retrieval_func):
        self.qa_func = qa_func
        self.retrieval_func = retrieval_func

    def _call(self, example) -> QuestionAnsweringResult:
        docs = list(make_docs(example))
        
        # First question
        question1 = example["question_decomposition"][0]["question"]
        query1 = question1
        docs1 = self.retrieval_func(docs, query1)
        context1 = "\n".join(doc['text'] for doc in docs1)
        result1 = self.qa_func(context=context1, question=question1)
        hop1 = {
            "question": question1,
            "query" : query1,
            "context": context1,
            "answer": result1.answer,
            "reasoning": result1.reasoning,
        }

        # Second question
        if result1.answer == "N/A":
            return {
                "answer": "N/A",
                "reasoning": result1.reasoning,
                "hops": [hop1],
            }

        question2 = example["question_decomposition"][1]["question"]
        question2 = question2.replace("#1", result1.answer)
        query2 = question2
        docs2 = self.retrieval_func(docs, query2)
        context2 = "\n".join(doc['text'] for doc in docs2)
        result2 = self.qa_func(context=context2, question=question2)
        hop2 = {
            "question": question2,
            "query": query2,
            "context": context2,
            "answer": result2.answer,
            "reasoning": result2.reasoning,
        }
        return QuestionAnsweringResult(answer=result2.answer, reasoning=result2.reasoning, raw_output=json.dumps([hop1, hop2]))

    def __call__(self, example, ignore_errors: bool = False) -> QuestionAnsweringResult:
        try:
            output = self._call(example)
        except Exception as exc:
            if ignore_errors:
                id = example['id']
                print(f"Failed to answer the question {id}\n{exc}")
                output = QuestionAnsweringResult(reasoning="", answer="N/A", raw_output=str(exc))
            else:
                raise
        return output

# %% ../../nbs/musique.baseline.ipynb 7
def benchmark(
    dataf: pd.DataFrame,
    pipeline: Callable,
    ignore_errors: bool = False,
    n_workers: int = 8,
) -> tuple[pd.DataFrame, dict]:

    def process(example):
        output = pipeline(example, ignore_errors=ignore_errors)
        example["predicted_answer"] = output.answer
        example["raw_llm_output"] = output
        return example

    rows = [] 
    with ThreadPoolExecutor(max_workers=n_workers) as executor:
        futures = [executor.submit(process, row) for _, row in dataf.iterrows()]
        for future in tqdm(as_completed(futures), total=len(dataf), desc="Processing samples"):
            rows.append(future.result())
    
    dataf = pd.DataFrame(rows)
    dataf = compare_answers(dataf)
    scores = calculate_metrics(dataf)
    scores["fuzzy_match"] = dataf["fuzzy_match"].mean()
    return dataf, scores



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/jerx.offline.llm.ipynb.

# %% auto 0
__all__ = ['log', 'parse_triplet_response', 'make_kg_triplet_extract_fn']

# %% ../../../nbs/jerx.offline.llm.ipynb 3
import json
from pathlib import Path
from ..utils import parse_triplets
from ...logging import get_logger

log = get_logger(__name__)

# %% ../../../nbs/jerx.offline.llm.ipynb 5
def parse_triplet_response(response: str, *args, **kwargs) -> list[tuple[str, str, str]]:
    triplets = parse_triplets(response.strip())
    return [(e1, rel, e2) if e1 != e2 else (e1, rel, e2 + "(obj)") for e1, rel, e2 in triplets]


def make_kg_triplet_extract_fn(inference_cache_filepath: Path):
    mapping = {}
    with open(inference_cache_filepath, "r") as f:
        for i, line in enumerate(f):
            record = json.loads(line.strip())
            mapping[record["text"]] = record["generation"]

    def extract_kg_triplets(text: str) -> list[tuple[str, str, str]]:
        return parse_triplet_response(mapping[text])

    return extract_kg_triplets



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/utils.ipynb.

# %% auto 0
__all__ = ['T', 'set_seed', 'most_common', 'context_chdir', 'generate_time_id', 'chunk_random', 'jprint', 'is_in', 'flatten_dict',
           'unflatten_dict', 'NestedDict', 'split_camel_case']

# %% ../nbs/utils.ipynb 3
import json
import os
import random
import re
from contextlib import contextmanager
from operator import eq
from pathlib import Path
from typing import Dict, List, Tuple, Union, Iterable, TypeVar, Generator

import numpy as np
import pandas as pd
from fastcore.basics import patch

# %% ../nbs/utils.ipynb 4
def set_seed(seed):
    np.random.seed(seed%(2**32-1))
    random.seed(seed)

# %% ../nbs/utils.ipynb 6
from collections import Counter

def most_common(lst):
    """returns the most common element of a collection"""
    return Counter(lst).most_common(1)[0][0]

# %% ../nbs/utils.ipynb 8
@patch
def ls_sorted(self:Path):
    "ls but sorts files by name numerically"
    return self.ls().sorted(key=lambda f: int(f.with_suffix('').name))

# %% ../nbs/utils.ipynb 9
# ref: https://dev.to/teckert/changing-directory-with-a-python-context-manager-2bj8
@contextmanager
def context_chdir(path: Union[Path, str]):
    """Sets the cwd within the context"""
    origin = Path().absolute()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(origin)

# %% ../nbs/utils.ipynb 11
from datetime import datetime

def generate_time_id(dt=None):
    """generates a string id from given datetime or now"""
    return (dt or datetime.now()).isoformat().rsplit('.', 1)[0].replace(':', '-')

# %% ../nbs/utils.ipynb 13
T = TypeVar("T")


def chunk_random(lst: List[T], min_chunk: int = 2, max_chunk: int = 4) -> Generator[List[T], None, None]:
    """
    Splits a list into random-sized chunks.

    Args:
        lst (list): The list to be split into chunks.
        min_chunk (int, optional): The minimum size of each chunk. Defaults to 2.
        max_chunk (int, optional): The maximum size of each chunk. Defaults to 4.

    Yields:
        list: A chunk of the original list.

    Returns:
        list: A list of chunks.

    """
    # Ensure the list has at least the minimum number of elements required for a chunk
    if len(lst) < min_chunk:
        return [lst]

    i = 0  # Initialize an index to traverse the list
    while i < len(lst):
        if len(lst) - i < min_chunk:
            break
        # Determine the size of the next chunk
        chunk_size = random.randint(min_chunk, min(max_chunk, len(lst) - i))
        # Add the chunk to the list of chunks
        yield lst[i : i + chunk_size]
        # Increment the index by the size of the chunk just added
        i += chunk_size


# %% ../nbs/utils.ipynb 15
def jprint(obj, indent=2, **kwargs):
    print(json.dumps(obj, indent=indent), **kwargs)

# %% ../nbs/utils.ipynb 17
def is_in(target, collection: Iterable, eq_fn=eq) -> bool:
    for item in collection:
        if eq_fn(item, target):
            return True
    return False

# %% ../nbs/utils.ipynb 19
def flatten_dict(d: Dict, sep='.') -> Dict:
    records = pd.json_normalize(d, sep=sep).to_dict(orient='records')
    if len(records):
        return records[0]
    return {}

def unflatten_dict(d: Dict, sep='.') -> Dict:
    res = {}
    for k, v in d.items():
        subkeys = k.split(sep)
        container = res
        for subkey in subkeys[:-1]:
            if subkey not in container:
                container[subkey] = {}
            container = container[subkey]
        container[subkeys[-1]] = v
    return res

# %% ../nbs/utils.ipynb 22
class NestedDict(dict):
    def __init__(self, data, sep='.'):
        super().__init__(data)
        self.sep = sep
    
    def at(self, keys: Union[str, List, Tuple], default=None):
        if isinstance(keys, str):
            keys = keys.split(self.sep)
        node = self
        for key in keys:
            if key not in node:
                return default
            node = node.get(key)
        return node

    def set(self, keys: Union[str, List, Tuple], value):
        if isinstance(keys, str):
            keys = keys.split(self.sep)
        node = self
        last_key = keys.pop()
        for key in keys:
            if key not in node:
                node[key] = dict()
            node = node[key]
        node[last_key] = value

    def flat(self) -> Dict:
        return flatten_dict(self, sep=self.sep)
    
    @classmethod
    def from_flat_dict(cls, data, sep='.'):
        return cls(unflatten_dict(data, sep=sep))
     

# %% ../nbs/utils.ipynb 25
def split_camel_case(input_str):
    # Use regular expression to find word boundaries in camel case
    matches = re.finditer('.+?(?:(?<=[a-z])(?=[A-Z])|(?<=[A-Z])(?=[A-Z][a-z])|$)', input_str)
    # Extract the words and return as a list
    return [m.group(0) for m in matches]



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/jerx.reward.qa.ipynb.

# %% auto 0
__all__ = ['log', 'DEFAULT_SYSTEM_PROMPT', 'USER_PROMPT', 'QuestionAnsweringResult', 'make_question_answer_func',
           'RewardAssessment', 'make_qa_reward_func']

# %% ../../../nbs/jerx.reward.qa.ipynb 3
import os

from openai import BadRequestError, OpenAI
from pydantic import BaseModel, Field

from ...logging import get_logger
from ...text.utils import fuzzy_match

log = get_logger(__name__)

# %% ../../../nbs/jerx.reward.qa.ipynb 5
import json

DEFAULT_SYSTEM_PROMPT = """You are an excellent Q&A system that is trusted around the world. You are given a question that requires multi-hop reasoning. Always answer the question using the provided context information, and not prior knowledge.

Some rules to follow:
1. Never directly reference the given context in your answer.
2. Avoid statements like 'Based on the context, ...' or 'The context information ...' or anything along those lines.

Output format:
Your output must be a single line in JSON such as:
{"reasoning": "Provide step by step multi-hop reasoning for the answer.", "answer": "Provide the final answer in 2-4 words."}
"""

USER_PROMPT = """The context information below is provided as a set of entity-relation-entity triplets from knowledge graph.
---------------------
{context}
---------------------
Given the context information and not prior knowledge, answer the question.
{question}
"""


class QuestionAnsweringResult(BaseModel):
    """Data model for answering the question."""

    reasoning: str = Field(description="Multi-hop reasoning for the answer.")
    answer: str = Field(description="The answer to the question in 2-4 words.")
    raw_output: str = Field(description="The raw output from the model.")


def make_question_answer_func(
    model_name: str = "gpt-3.5-turbo",
    client: OpenAI = None,
    completion_kwargs: dict | None = None,
):
    if client is None:
        client = OpenAI()

    if completion_kwargs is None:
        completion_kwargs = {}

    def func(context: str, question: str) -> QuestionAnsweringResult:
        messages = [
            {
                "role": "system",
                "content": DEFAULT_SYSTEM_PROMPT,
            },
            {
                "role": "user",
                "content": USER_PROMPT.format(context=context, question=question),
            },
        ]
        chat_completion = client.chat.completions.create(
            model=model_name,
            messages=messages,
            **completion_kwargs,
        )
        text = chat_completion.choices[0].message.content
        try:
            output = json.loads(text)
        except json.JSONDecodeError:
            log.error("Failed to decode the JSON output: %s", text)
            output = {}
        return QuestionAnsweringResult(
            answer=output.get("answer", "N/A"),
            reasoning=output.get("reasoning", ""),
            raw_output=text,
        )

    return func

# %% ../../../nbs/jerx.reward.qa.ipynb 6
class RewardAssessment(QuestionAnsweringResult):
    reward: float = Field(description="The reward value for the answer.")

def make_qa_reward_func(model_name: str = "gpt-3.5-turbo", answer_comparator=fuzzy_match, completion_kwargs: dict | None = None):
    qa = make_question_answer_func(model_name, completion_kwargs=completion_kwargs)

    def reward(context: str, question: str, answers: list[str]) -> RewardAssessment:
        try:
            qa_result = qa(context, question)
        except BadRequestError as e:
            log.warning(f"Failed to assess generation: {e}")
            return RewardAssessment(answer="", reasoning=str(e), reward=0.0)
        correct = any(answer_comparator(qa_result.answer, answer) for answer in answers)
        reward = 1.0 if correct else 0.0
        return RewardAssessment(**qa_result.dict(), reward=reward)

    return reward



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/mhqa.llm.ipynb.

# %% auto 0
__all__ = ['log', 'DEFAULT_SYSTEM_PROMPT', 'USER_PROMPT', 'make_mhqa_chat', 'parse_llm_generation', 'QuestionAnsweringResult',
           'make_question_answer_func']

# %% ../../nbs/mhqa.llm.ipynb 4
import json
import os

from openai import OpenAI
from pydantic import BaseModel, Field

from ..logging import get_logger

log = get_logger(__name__)

# %% ../../nbs/mhqa.llm.ipynb 5
DEFAULT_SYSTEM_PROMPT = """You are an excellent Q&A system that is trusted around the world. You are given a question that requires multi-hop reasoning. Always answer the question using the provided context information, and not prior knowledge.

Some rules to follow:
1. Never directly reference the given context in your answer.
2. Avoid statements like 'Based on the context, ...' or 'The context information ...' or anything along those lines.

Output format:
Your output must be a single line in JSON such as:
{"reasoning": "Provide step by step multi-hop reasoning for the answer.", "answer": "Provide the final answer in 2-4 words."}
"""

USER_PROMPT = """The context information below is provided as a set of entity-relation-entity triplets from knowledge graph.
---------------------
{context}
---------------------
Given the context information and not prior knowledge, answer the question.
{question}
"""

def make_mhqa_chat(context: str, question: str) -> list[dict]:
    return [
            {
                "role": "system",
                "content": DEFAULT_SYSTEM_PROMPT,
            },
            {
                "role": "user",
                "content": USER_PROMPT.format(context=context, question=question),
            },
        ]

# %% ../../nbs/mhqa.llm.ipynb 6
def parse_llm_generation(output: str):
    return json.loads(output)

# %% ../../nbs/mhqa.llm.ipynb 7
class QuestionAnsweringResult(BaseModel):
    """Data model for answering the question."""

    reasoning: str = Field(description="Multi-hop reasoning for the answer.")
    answer: str = Field(description="The answer to the question in 2-4 words.")
    raw_output: str = Field(description="The raw output from the model.")


def make_question_answer_func(
    model_name: str = "gpt-3.5-turbo",
    client: OpenAI = None,
    completion_kwargs: dict | None = None,
):
    if client is None:
        client = OpenAI()

    if completion_kwargs is None:
        completion_kwargs = {}

    def func(context: str, question: str) -> QuestionAnsweringResult:
        messages = make_mhqa_chat(context, question)
        chat_completion = client.chat.completions.create(
            model=model_name,
            messages=messages,
            **completion_kwargs,
        )
        text = chat_completion.choices[0].message.content
        output = parse_llm_generation(text)
        return QuestionAnsweringResult(answer=output["answer"], reasoning=output["reasoning"], raw_output=text)

    return func



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/hf.transformers.llama3.ipynb.

# %% auto 0
__all__ = ['prepare_llama3_for_training', 'prepare_llama3_for_inference']

# %% ../../../nbs/hf.transformers.llama3.ipynb 3
def prepare_llama3_for_training(tokenizer, model):
    tokenizer.pad_token = tokenizer.eos_token
    tokenizer.pad_token_id = tokenizer.eos_token_id
    tokenizer.padding_side = "right" # Fix weird overflow issue with fp16 training
    model.config.pretraining_tp = 1
    model.config.use_cache = False

def prepare_llama3_for_inference(tokenizer, model):
    tokenizer.pad_token = tokenizer.eos_token
    tokenizer.pad_token_id = tokenizer.eos_token_id
    tokenizer.padding_side = "left" # During inference, the padding must be on the left for decoder-only models



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/hf.transformers.generation.ipynb.

# %% auto 0
__all__ = ['log', 'flat_pipeline', 'preprocess_generation_params', 'generate']

# %% ../../../nbs/hf.transformers.generation.ipynb 3
from copy import deepcopy

from tqdm import tqdm

from ...logging import get_logger

log = get_logger(__name__)

# %% ../../../nbs/hf.transformers.generation.ipynb 4
def flat_pipeline(pipe):
    def func(inputs, **kwargs) -> list[str]:
        return [result[0]["generated_text"] for result in tqdm(pipe(inputs, **kwargs), total=len(inputs))]

    return func


def preprocess_generation_params(tokenizer, generation_params: dict) -> dict:
    generation_params = deepcopy(generation_params)

    terminators = generation_params.pop("terminators", [])
    eos_token_ids = {tokenizer.eos_token_id}
    for terminator in terminators:
        if isinstance(terminator, int):
            eos_token_ids.add(terminator)
        elif isinstance(terminator, str):
            eos_token_ids.add(tokenizer.convert_tokens_to_ids(terminator))
        else:
            raise ValueError(f"Invalid terminator token {terminator}.")
    generation_params["eos_token_id"] = sorted(eos_token_ids)

    generation_params["pad_token_id"] = tokenizer.eos_token_id

    return generation_params


def generate(
    pipe,
    inputs,
    **generation_kwargs,
) -> list[str]:
    # Setup generation parameters
    generation_kwargs = preprocess_generation_params(pipe.tokenizer, generation_kwargs)
    generation_kwargs["return_full_text"] = False

    # Generate text
    log.info(f"Running pipeline on {len(inputs)} samples...")
    return flat_pipeline(pipe)(inputs, **generation_kwargs)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/jerx.reward.llm.ipynb.

# %% auto 0
__all__ = ['log', 'SYSTEM_PROMPT', 'Assessment', 'assess_quality']

# %% ../../../nbs/jerx.reward.llm.ipynb 3
import magentic
from pydantic import BaseModel, Field

from ...logging import get_logger

log = get_logger(__name__)

# %% ../../../nbs/jerx.reward.llm.ipynb 5
SYSTEM_PROMPT = """
Assess the quality of the entity-relation-entity triplets extracted from the document. Provide a brief explanation for your assessment and a score between 0 and 1. The score should reflect the quality of the triplets in terms of their relevance and completeness. The explanation should describe the strengths and weaknesses of the triplets and how they could be improved. Your output must 
Reasoning: 1-line explanation of the assessment.
Score: a number between 0 and 1.
""".strip()

@magentic.chatprompt(
        magentic.SystemMessage(SYSTEM_PROMPT),
        magentic.UserMessage("""
# Document
Washington Township is one of the thirteen townships of Henry County, Ohio, United States. As of the 2010
census the population was 1,912, of whom 1,794 lived in the unincorporated portion of the township.
# Triplets
Washington Township | is one of | thirteen townships of Henry County
Washington Township | located in | Henry County, Ohio
Washington Township | had population in 2010 census | 1,912
Unincorporated portion of Washington Township | had population in 2010 census | 1,794
Washington Township | part of census | 2010
""".strip()),
        magentic.AssistantMessage("""
Reasoning: The triplets are well-crafted and highly relevant but miss some opportunities for providing comprehensive details that would make them even more useful for answering a broader range of related questions.
Score: 0.6
""".strip()),
        magentic.UserMessage("""
# Document
Turkeycock Mountain is a mountain summit located in Franklin County, Virginia and Henry County, Virginia.
Rising out of the eastern foothills of the Blue Ridge Mountains, Turkeycock Mountain rises to above sea level and is
located at . A portion of the mountain is protected as the Turkeycock Wildlife Management Area, which is open to the
public.

# Triplets
Turkeycock Mountain | located in | Franklin County, Virginia
Turkeycock Mountain | located in | Henry County, Virginia
Turkeycock Mountain | part of | Blue Ridge Mountains
Turkeycock Mountain | feature | rises to above sea level
Turkeycock Mountain | rises from | eastern foothills of Blue Ridge Mountains
Turkeycock Mountain | elevation | above sea level
Turkeycock Wildlife Management Area | part of | Turkeycock Mountain
Turkeycock Wildlife Management Area | status | protected
Turkeycock Wildlife Management Area | open to | public
Turkeycock Wildlife Management Area | allows | public access
""".strip()),
        magentic.AssistantMessage("""
Reasoning: The triplets are accurate and complete.
Score: 1.0
""".strip()),
        magentic.UserMessage("""
# Document
Joachim II ( or "Hektor"; 13 January 1505 – 3 January 1571) was a Prince-elector of the Margraviate of Brandenburg (1535–1571), the sixth member of the House of Hohenzollern. Joachim II was the eldest son of Joachim I Nestor, Elector of Brandenburg and his wife Elizabeth of Denmark, Norway, and Sweden. He received the cognomen Hector after the Trojan prince and warrior for his qualities and prowess.

# Triplets
Joachim II | legacy | helped to establish the Hohenzollern dynasty as a major power in Europe
Joachim II | legacy | helped to establish the Hohenzollern dynasty as a major power in Europe
Joachim II | legacy | helped to establish the Hohenzollern dynasty as a major power in Europe
Joachim II | legacy | helped to establish the Hohenzollern dynasty as a major power in Europe
Joachim II | legacy | helped to establish the Hohenzollern dynasty as a major power in the world
Joachim II | legacy | helped to establish the Hohenzollern dynasty as a major power in the world
Joachim II | legacy | helped to establish the Hohenzollern dynasty as a major power in the world
Joachim II | legacy | helped to establish the Hohenzollern dynasty as a major power in the world
Joachim II | legacy | helped to establish the Hohenzollern dynasty as a major power in the world
Joachim II | legacy | helped to establish the Hohenzollern dynasty as a major power in the world
Joachim II | legacy | helped to establish the Hohenzollern dynasty as a major power in the world
Joachim II | legacy | helped to establish the Hohenzollern dynasty as a major power in the world
Joachim II | legacy | helped to establish the Hohenzollern dynasty as a major power in the world
""".strip()),
        magentic.AssistantMessage("""
Reasoning: There are many duplicate triplets in the response and the triplets do not capture the full range of information in the document. The format of each triplet is correct.
Score: 0.05
""".strip()),

        magentic.UserMessage("""
# Document
{document}

# Triplets
{triplets_str}
""".strip()),
)
def _assess_triplets(document: str, triplets_str: str) -> str: ...


# %% ../../../nbs/jerx.reward.llm.ipynb 6
class Assessment(BaseModel):
    reasoning: str = Field(description="A brief reasoning for the assessment.")
    reward: float = Field(ge=0, le=1, description="The reward of the assessment.")

def assess_quality(document: str, triplets_str: str) -> Assessment:
    result = _assess_triplets(document, triplets_str)
    lines = result.splitlines()
    if len(lines) < 2:
        return Assessment(reasoning=result, reward=0.0)
    
    last_line = lines[-1]
    if not last_line.startswith("Score: "):
        return Assessment(reasoning=result, reward=0.0)
    
    reward = float(last_line[len("Score:"):].strip())
    return Assessment(reasoning="\n".join(lines[:-1]), reward=reward)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/musique.qa.ipynb.

# %% auto 0
__all__ = ['log', 'USER_PROMPT', 'EXAMPLE_CONTEXT', 'EXAMPLE_QUESTION', 'SYSTEM_PROMPT_STANDARD', 'SYSTEM_PROMPT_COT',
           'SYSTEM_PROMPT_COT_FS', 'EXAMPLE_COT_RESPONSE', 'SYSTEM_PROMPT_CTE', 'EXAMPLE_CTE_RESPONSE',
           'answer_question_standard', 'answer_question_cot', 'answer_question_cot_fs', 'answer_question_cte']

# %% ../../nbs/musique.qa.ipynb 4
import magentic
import openai
from pydantic import BaseModel, Field

from ..logging import get_logger

log = get_logger(__name__)

# %% ../../nbs/musique.qa.ipynb 6
USER_PROMPT = """The context information is provided below.
---------------------
{context}
---------------------
Given the context information and not prior knowledge, answer the question.
{question}
"""

# %% ../../nbs/musique.qa.ipynb 7
EXAMPLE_CONTEXT = """
Glenhis Hernández (born 7 October 1990 in Havana) is a taekwondo practitioner from Cuba. She was the 2013 World
Champion in middleweight.

The current mayor of Havana ("President of the People's Power Provincial Assembly") is Marta Hernández Romero, she
was elected on March 5, 2011.
""".strip()

EXAMPLE_QUESTION = "Who is the current mayor of the city Glenhis Hernández was born?"


# %% ../../nbs/musique.qa.ipynb 9
SYSTEM_PROMPT_STANDARD = """
You are an excellent question-answering system known for providing accurate and reliable answers. Your responses should be solely based on the context information given, without drawing on prior knowledge. 

# Output format
Answer: [answer in 2-4 words]
""".strip()

def answer_question_standard(
    context: str,
    question: str,
    model_name: str = "gpt-3.5-turbo",
    completion_kwargs: dict | None = None,
    client = None
) -> dict:
    
    if client is None:
        client = openai.Client()
    
    completion_kwargs = completion_kwargs or {}
    messages = [
        {
            "role": "system",
            "content": SYSTEM_PROMPT_STANDARD,
        },
        {
            "role": "user",
            "content": USER_PROMPT.format(context=context, question=question),
        },
    ]
    chat_completion = client.chat.completions.create(
            model=model_name,
            messages=messages,
            **completion_kwargs,
        )
    generation = chat_completion.choices[0].message.content
    if "Answer:" not in generation:
        return dict(answer="", generation=generation)
    answer = generation.split("Answer: ")[1].strip()
    return dict(answer=answer, generation=generation)

# %% ../../nbs/musique.qa.ipynb 12
SYSTEM_PROMPT_COT = """
You are an excellent question-answering system known for providing accurate and reliable answers. Your responses should be solely based on the context information given, without drawing on prior knowledge. Always provide clear and logical step-by-step reasoning in your response.

# Output format
Reasoning: [Step-by-step reasoning for the answer.]
Answer: [answer in 2-4 words]
""".strip()

def answer_question_cot(
    context: str,
    question: str,
    model_name: str = "gpt-3.5-turbo",
    completion_kwargs: dict | None = None,
    client = None
) -> dict:

    if client is None:
        client = openai.Client()
    
    completion_kwargs = completion_kwargs or {}
    messages = [
        {
            "role": "system",
            "content": SYSTEM_PROMPT_COT,
        },
        {
            "role": "user",
            "content": USER_PROMPT.format(context=context, question=question),
        },
    ]
    chat_completion = client.chat.completions.create(
            model=model_name,
            messages=messages,
            **completion_kwargs,
        )
    generation = chat_completion.choices[0].message.content
    # Parse the response
    answer = ""
    reasoning = ""
    for line in generation.splitlines():
        if line.startswith("Answer:"):
            answer = line.split("Answer:")[1].strip()
        else:
            reasoning += line.replace("Reasoning:", "").strip()
    return dict(reasoning=reasoning, answer=answer, generation=generation)

# %% ../../nbs/musique.qa.ipynb 14
SYSTEM_PROMPT_COT_FS = """You are an excellent question-answering system known for providing accurate and reliable answers. Your responses should be solely based on the context information given, without drawing on prior knowledge. Always provide clear and logical step-by-step reasoning in your response.

# Output format
Reasoning: [Step-by-step reasoning for the answer.]
Answer: [answer in 2-4 words]
"""

EXAMPLE_COT_RESPONSE = """
Reasoning:
1. The context states that Glenhis Hernández was born in Havana.
2. It also mentions that the current mayor (President of the People's Power Provincial Assembly) of Havana is Marta Hernández Romero.
3. Marta Hernández Romero has been the mayor since her election on March 5, 2011.

Answer: Marta Hernández Romero
""".strip()

def answer_question_cot_fs(
    context: str,
    question: str,
    model_name: str = "gpt-3.5-turbo",
    completion_kwargs: dict | None = None,
    client = None
) -> dict:

    if client is None:
        client = openai.Client()
    
    completion_kwargs = completion_kwargs or {}
    messages = [
        {
            "role": "system",
            "content": SYSTEM_PROMPT_COT_FS,
        },
        {
            "role": "user",
            "content": USER_PROMPT.format(context=EXAMPLE_CONTEXT, question=EXAMPLE_QUESTION),
        },
        {
            "role": "assistant",
            "content": EXAMPLE_COT_RESPONSE
        },
        {
            "role": "user",
            "content": USER_PROMPT.format(context=context, question=question),
        },
    ]
    chat_completion = client.chat.completions.create(
            model=model_name,
            messages=messages,
            **completion_kwargs,
        )
    generation = chat_completion.choices[0].message.content
    # Parse the response
    answer = ""
    reasoning = ""
    for line in generation.splitlines():
        if line.startswith("Answer:"):
            answer = line.split("Answer:")[1].strip()
        else:
            reasoning += line.replace("Reasoning:", "") + "\n"
    return dict(reasoning=reasoning.strip(), answer=answer, generation=generation)

# %% ../../nbs/musique.qa.ipynb 17
SYSTEM_PROMPT_CTE = """You are an excellent question-answering system known for providing accurate and reliable answers. Your responses should be solely based on the context information given, without drawing on prior knowledge.

Before answering the question, first, you extract relevant entity-relation-entity triplets from the context. Then, you answer the question based on the triplets. 

# Output format
Triplets: [A list of entity-relation-entity triplets extracted from the context.]
Answer: [answer in 2-4 words]
"""

EXAMPLE_CTE_RESPONSE = """
Triplets: 
Glenhis Hernández | birth place | Havana
Marta Hernández Romero | serves as | mayor of Havana

Answer: Marta Hernández Romero
""".strip()

def answer_question_cte(
    context: str,
    question: str,
    model_name: str = "gpt-3.5-turbo",
    completion_kwargs: dict | None = None,
    client = None
) -> dict:

    if client is None:
        client = openai.Client()
    
    completion_kwargs = completion_kwargs or {}
    messages = [
        {
            "role": "system",
            "content": SYSTEM_PROMPT_CTE,
        },
        {
            "role": "user",
            "content": USER_PROMPT.format(context=EXAMPLE_CONTEXT, question=EXAMPLE_QUESTION),
        },
        {
            "role": "assistant",
            "content": EXAMPLE_CTE_RESPONSE
        },
        {
            "role": "user",
            "content": USER_PROMPT.format(context=context, question=question),
        },
    ]
    chat_completion = client.chat.completions.create(
            model=model_name,
            messages=messages,
            **completion_kwargs,
        )
    generation = chat_completion.choices[0].message.content
    # Parse the response
    answer = ""
    triplets = []
    for line in generation.splitlines():
        if line.startswith("Answer:"):
            answer = line.split("Answer:")[1].strip()
        elif "|" in line:
            triplets.append(line.strip())
    return dict(triplets=triplets, answer=answer, generation=generation)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/qdecomp.llm.ipynb.

# %% auto 0
__all__ = ['QUESTION_DECOMPOSITION_SYSTEM_PROMPT_TEMPLATE', 'parse_llm_generation', 'make_question_decomposer']

# %% ../../nbs/qdecomp.llm.ipynb 3
import re
from openai import OpenAI

# %% ../../nbs/qdecomp.llm.ipynb 4
QUESTION_DECOMPOSITION_SYSTEM_PROMPT_TEMPLATE = """Decompose the given question into n sub-questions such that when sub-questions are answered, the original question can be answered correctly.
The second subquestion must refer to the answer of the first question by `#1` as in the examples below. Do not create open-ended sub-questions like "Who is ..." or "How is ...".

Avoid statements like 'Here are the decomposed sub-question, ...' or 'Sure! Here are..' or anything along those lines. Just provide sub-questions as illustrated in the examples below.

Question: Who founded the journal that published The Review of Communication?
Sub-questions:
1. Who published The Review of Communication?
2. Who founded #1?

Question: When was the institute that owned The Collegian founded?
Sub-questions:
1. Which institute does own The Collegian?
2. When #1 founded?

Question: {question}
Sub-questions:
"""

# %% ../../nbs/qdecomp.llm.ipynb 5
def parse_llm_generation(output: str):
    if output.startswith("Sub-questions:"):
        output = output.split("Sub-questions:", 1)[-1]
    enumeration_pattern = r'^\d+\.\s'
    for line in output.splitlines():
        # check if the line starts with enumeration
        if not re.match(enumeration_pattern, line):
            continue
        yield line.split(".", 1)[-1].strip()

# %% ../../nbs/qdecomp.llm.ipynb 6
def make_question_decomposer(
    model: str = "gpt-3.5-turbo",
    client: OpenAI = None,
    completion_kwargs: dict | None = None,
):
    if client is None:
        client = OpenAI()

    if completion_kwargs is None:
        completion_kwargs = {}

    def func(question: str) -> list[str]:
        messages = [
            {
                "role": "system",
                "content": QUESTION_DECOMPOSITION_SYSTEM_PROMPT_TEMPLATE.format(question=question),
            },
        ]
        chat_completion = client.chat.completions.create(
            model=model,
            messages=messages,
            **completion_kwargs,
        )
        text = chat_completion.choices[0].message.content
        return list(parse_llm_generation(text))

    return func



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/hf.transformers.utils.ipynb.

# %% auto 0
__all__ = ['log', 'prepare_model_kwargs', 'load_tokenizer_model', 'merge_adapters_and_publish']

# %% ../../../nbs/hf.transformers.utils.ipynb 3
from copy import deepcopy
import torch
from transformers import AutoTokenizer, AutoModelForCausalLM
from transformers.utils import is_flash_attn_2_available
from ...logging import get_logger

log = get_logger(__name__)

# %% ../../../nbs/hf.transformers.utils.ipynb 4
def prepare_model_kwargs(model_kwargs: dict) -> dict:
    model_kwargs = deepcopy(model_kwargs)
    
    # Setup quantization config 
    if (quantization_config := model_kwargs.get("quantization_config")) and isinstance(quantization_config, dict):
        from transformers import BitsAndBytesConfig

        model_kwargs["quantization_config"] = BitsAndBytesConfig(**quantization_config)
    
    # Setup torch dtype
    if (torch_dtype := model_kwargs.get("torch_dtype")) and (torch_dtype != "auto"):
        model_kwargs["torch_dtype"] = getattr(torch, torch_dtype)

    # Setup device map
    if "device_map" not in model_kwargs:
        model_kwargs["device_map"] = {"": torch.cuda.current_device()}

    # Setup flash attention 2
    if model_kwargs.get("attn_implementation") == "flash_attention_2":
        if not is_flash_attn_2_available():
            log.warning("Flash attention 2 is not available. Falling back to default attention.")
            del model_kwargs["attn_implementation"]
    return model_kwargs

# %% ../../../nbs/hf.transformers.utils.ipynb 5
def load_tokenizer_model(
    model_name_or_path: str,
    *,
    auto_model_cls=None,
    **model_kwargs,
):

    model_kwargs = prepare_model_kwargs(model_kwargs)

    if auto_model_cls is None:
        if "-peft" in model_name_or_path:
            from peft import AutoPeftModelForCausalLM

            auto_model_cls = AutoPeftModelForCausalLM
        else:
            auto_model_cls = AutoModelForCausalLM

    # Load model
    model = auto_model_cls.from_pretrained(
        model_name_or_path,
        **model_kwargs,
    )
    # Load tokenizer
    if auto_model_cls == AutoModelForCausalLM:
        tokenizer_id = model_name_or_path
    else:
        from peft import AutoPeftModelForCausalLM

        if auto_model_cls == AutoPeftModelForCausalLM:
            tokenizer_id = model.active_peft_config.base_model_name_or_path
        else:
            raise ValueError(f"Unknown auto_model_cls: {auto_model_cls}")
    tokenizer = AutoTokenizer.from_pretrained(tokenizer_id, trust_remote_code=True)
    return tokenizer, model

# %% ../../../nbs/hf.transformers.utils.ipynb 6
def merge_adapters_and_publish(
    model_id: str,
    torch_dtype=torch.float16,
    device_map={"": 0},
    merged_model_id: str=None,
):
    from peft import AutoPeftModelForCausalLM
    
    if isinstance(torch_dtype, str) and torch_dtype != "auto":
        torch_dtype = getattr(torch, torch_dtype) 

    log.info(f"Loading model and tokenizer for {model_id}")
    model = AutoPeftModelForCausalLM.from_pretrained(
        model_id,
        torch_dtype=torch_dtype,
        device_map=device_map,
    )
    tokenizer = AutoTokenizer.from_pretrained(model_id, trust_remote_code=True)

    log.info("Merging adapters to model...")
    model = model.merge_and_unload()

    if merged_model_id is None:
        merged_model_id = f"{model_id}-merged"
    log.info(f"Pushing merged model to HF hub as {merged_model_id}")
    model.push_to_hub(merged_model_id)
    tokenizer.push_to_hub(merged_model_id)
    return merged_model_id



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/hf.datasets.utils.ipynb.

# %% auto 0
__all__ = ['log', 'load_datasets', 'concatenate_dataset_dicts', 'chunk_random_dataset']

# %% ../../../nbs/hf.datasets.utils.ipynb 3
from datasets import concatenate_datasets, DatasetDict, load_dataset, Dataset

from ...logging import get_logger
from ...utils import chunk_random

log = get_logger(__name__)

# %% ../../../nbs/hf.datasets.utils.ipynb 4
def load_datasets(dataset_kwargs_list: list[dict]) -> Dataset:
    datasets = []
    for dataset_kwargs in dataset_kwargs_list:
        dataset = load_dataset(**dataset_kwargs)
        datasets.append(dataset)
    return concatenate_datasets(datasets)

# %% ../../../nbs/hf.datasets.utils.ipynb 5
def concatenate_dataset_dicts(dataset_dicts: list[DatasetDict]):
    """
    Concatenate multiple `DatasetDict` objects into a single `DatasetDict`.
    """
    splits = {key for dd in dataset_dicts for key in dd.keys()}
    output_dsd = {}
    for split in splits:
        dataset_list = []
        for dd in dataset_dicts:
            if split in dd.keys():
                dataset_list.append(dd[split])
        output_dsd[split] = concatenate_datasets(dataset_list)
    return DatasetDict(output_dsd)


# %% ../../../nbs/hf.datasets.utils.ipynb 7
def chunk_random_dataset(ds, min_chunk=1, max_chunk=3):
    for indices in chunk_random(range(len(ds)), min_chunk, max_chunk):
        yield ds.select(indices).to_list()



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/qa.ablation.ipynb.

# %% auto 0
__all__ = ['log', 'USER_PROMPT', 'SYSTEM_PROMPT_STANDARD', 'SYSTEM_PROMPT_COT', 'SYSTEM_PROMPT_COT_FS', 'SYSTEM_PROMPT_CTE',
           'SYSTEM_PROMPT_CTE_COT', 'QuestionAnsweringResultStandard', 'answer_question_standard',
           'QuestionAnsweringResultCOT', 'answer_question_cot', 'QuestionAnsweringResultCOTFS',
           'answer_question_cot_fs', 'QuestionAnsweringResultCTE', 'answer_question_cte',
           'QuestionAnsweringResultCTECOT', 'answer_question_cte_cot']

# %% ../../nbs/qa.ablation.ipynb 4
import magentic
from pydantic import BaseModel, Field

from ..logging import get_logger

log = get_logger(__name__)

# %% ../../nbs/qa.ablation.ipynb 5
USER_PROMPT = """The context information is provided below.
---------------------
{context}
---------------------
Given the context information and not prior knowledge, answer the question.
{question}
"""

# %% ../../nbs/qa.ablation.ipynb 6
SYSTEM_PROMPT_STANDARD = """You are an excellent question-answering system known for providing accurate and reliable answers. Your responses should be solely based on the context information given, without drawing on prior knowledge."""

class QuestionAnsweringResultStandard(BaseModel):
    """Data model for answering the question."""

    answer: str = Field(description="The answer to the question in 2-4 words.")


@magentic.chatprompt(
    magentic.SystemMessage(SYSTEM_PROMPT_STANDARD), 
    magentic.UserMessage(USER_PROMPT),
)
def answer_question_standard(
    context: str,
    question: str,
) -> QuestionAnsweringResultStandard: ...

# %% ../../nbs/qa.ablation.ipynb 7
SYSTEM_PROMPT_COT = """You are an excellent question-answering system known for providing accurate and reliable answers. Your responses should be solely based on the context information given, without drawing on prior knowledge. Always provide clear and logical step-by-step reasoning in your response."""

class QuestionAnsweringResultCOT(BaseModel):
    """Data model for answering the question."""

    reasoning: str = Field(description="Step-by-step reasoning for the answer.")
    answer: str = Field(description="The answer to the question in 2-4 words.")


@magentic.chatprompt(
    magentic.SystemMessage(SYSTEM_PROMPT_COT), 
    magentic.UserMessage(USER_PROMPT),
)
def answer_question_cot(
    context: str,
    question: str,
) -> QuestionAnsweringResultCOT: ...

# %% ../../nbs/qa.ablation.ipynb 8
SYSTEM_PROMPT_COT_FS = """You are an excellent question-answering system known for providing accurate and reliable answers. Your responses should be solely based on the context information given, without drawing on prior knowledge. Always provide clear and logical step-by-step reasoning in your response.

# Example
Context: "Glenhis Hernández (born 7 October 1990 in Havana) is a taekwondo practitioner from Cuba. She was the 2013 World
Champion in middleweight.

The current mayor of Havana ("President of the People's Power Provincial Assembly") is Marta Hernández Romero, she
was elected on March 5, 2011."

Question: "Who is the current mayor of the city Glenhis Hernández was born?"

Reasoning:
1. The context states that Glenhis Hernández was born in Havana.
2. It also mentions that the current mayor (President of the People's Power Provincial Assembly) of Havana is Marta Hernández Romero.
3. Marta Hernández Romero has been the mayor since her election on March 5, 2011.

Answer: "Marta Hernández Romero"
"""

class QuestionAnsweringResultCOTFS(BaseModel):
    """Data model for answering the question."""

    reasoning: str = Field(description="Step-by-step reasoning for the answer.")
    answer: str = Field(description="The answer to the question in 2-4 words.")


@magentic.chatprompt(
    magentic.SystemMessage(SYSTEM_PROMPT_COT_FS), 
    magentic.UserMessage(USER_PROMPT),
)
def answer_question_cot_fs(
    context: str,
    question: str,
) -> QuestionAnsweringResultCOTFS: ...

# %% ../../nbs/qa.ablation.ipynb 9
SYSTEM_PROMPT_CTE = """You are an excellent question-answering system known for providing accurate and reliable answers. Your responses should be solely based on the context information given, without drawing on prior knowledge.

Before answering the question, first, you extract relevant entity-relation-entity triplets from the context. Then, you answer the question based on the triplets. For instance, 

# Example
Context: "Glenhis Hernández (born 7 October 1990 in Havana) is a taekwondo practitioner from Cuba. She was the 2013 World
Champion in middleweight.

The current mayor of Havana ("President of the People's Power Provincial Assembly") is Marta Hernández Romero, she
was elected on March 5, 2011."

Question: "Who is the current mayor of city Glenhis Hernández was born?"

Triplets: 
Glenhis Hernández | birth place | Havana
Marta Hernández Romero | serves as | mayor of Havana

Answer: "Marta Hernández Romero"
"""

class _QuestionAnsweringResultCTE(BaseModel):
    """Data model for answering the question."""

    triplets: list[str] = Field(description="A list of entity-relation-entity triplets extracted from the context.")
    answer: str = Field(description="The answer to the question in 2-4 words.")

class QuestionAnsweringResultCTE(_QuestionAnsweringResultCTE):
    reasoning: str = ""

@magentic.chatprompt(
    magentic.SystemMessage(SYSTEM_PROMPT_CTE), 
    magentic.UserMessage(USER_PROMPT),
)
def _answer_question_cte(
    context: str,
    question: str,
) -> _QuestionAnsweringResultCTE: ...

def answer_question_cte(
    context: str,
    question: str,
) -> QuestionAnsweringResultCTE: 
    result = _answer_question_cte(context, question)
    return QuestionAnsweringResultCTE(triplets=result.triplets, answer=result.answer)

# %% ../../nbs/qa.ablation.ipynb 10
SYSTEM_PROMPT_CTE_COT = """You are an excellent question-answering system known for providing accurate and reliable answers. Your responses should be solely based on context the information given, without drawing on prior knowledge. Always provide clear and logical step-by-step reasoning in your response.

Before answering the question, first, you extract relevant entity-relation-entity triplets from the context. Then, you answer the question based on the triplets. For instance, 

# Example
Context: "Glenhis Hernández (born 7 October 1990 in Havana) is a taekwondo practitioner from Cuba. She was the 2013 World
Champion in middleweight.

The current mayor of Havana ("President of the People's Power Provincial Assembly") is Marta Hernández Romero, she
was elected on March 5, 2011."

Question: "Who is the current mayor of city Glenhis Hernández was born?"

Triplets: "Glenhis Hernández (Athlete) | born on | October 7, 1990
Glenhis Hernández (Athlete) | birth place | Havana
Glenhis Hernández (Athlete) | specializes in | taekwondo
Glenhis Hernández (Athlete) | won | 2013 World Champion title (Middleweight)
Marta Hernández Romero (Politician) | serves as | mayor of Havana
Marta Hernández Romero (Politician) | holds | the position of 'President of the People's Power Provincial Assembly'
Marta Hernández Romero (Politician) | elected on | March 5, 2011."

Reasoning: "Glenhis Hernández (Athlete) | birth place | Havana
This indicates that Glenhis Hernández was born in Havana.

Marta Hernández Romero (Politician) | serves as | mayor of Havana
This states that Marta Hernández Romero is the mayor of Havana.

From these triplets, we conclude that Marta Hernández Romero is the mayor of Havana."

Answer: "Marta Hernández Romero"
"""

class QuestionAnsweringResultCTECOT(BaseModel):
    """Data model for answering the question."""

    triplets: list[str] = Field(description="A list of entity-relation-entity triplets extracted from the context.")
    reasoning: str = Field(description="Step-by-step reasoning for the answer.")
    answer: str = Field(description="The answer to the question in 2-4 words.")


@magentic.chatprompt(
    magentic.SystemMessage(SYSTEM_PROMPT_CTE_COT), 
    magentic.UserMessage(USER_PROMPT),
)
def answer_question_cte_cot(
    context: str,
    question: str,
) -> QuestionAnsweringResultCTECOT: ...



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/musique.multihopjerx.ipynb.

# %% auto 0
__all__ = ['DEFAULT_SYSTEM_PROMPT', 'USER_PROMPT', 'make_docs', 'make_qa_chat', 'parse_llm_generation', 'QuestionAnsweringResult',
           'make_question_answer_func', 'BaselineMultiHop', 'benchmark']

# %% ../../nbs/musique.multihopjerx.ipynb 3
import json
from typing import Callable

import pandas as pd
from tqdm.auto import tqdm

from ..jerx.reward.llm import QuestionAnsweringResult
from .eval import calculate_metrics, compare_answers

tqdm.pandas()

# %% ../../nbs/musique.multihopjerx.ipynb 4
def make_docs(example):
    ps = example["paragraphs"]
    for p in ps:
        idx = p["idx"]
        title = p["title"]
        body = p["paragraph_text"]
        is_supporting = p["is_supporting"]
        text = f"# {title}\n{body}"
        yield dict(
            text=text,
            is_supporting=is_supporting,
            parent_id=example["id"],
            idx=idx,
        )

# %% ../../nbs/musique.multihopjerx.ipynb 5
# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/qa.llm.ipynb.

import json
import os

from openai import OpenAI
from pydantic import BaseModel, Field

DEFAULT_SYSTEM_PROMPT = """You are an excellent Q&A system that is trusted around the world. Always answer the question using the provided context information, and not prior knowledge.

Some rules to follow:
1. Never directly reference the given context in your answer.
2. Avoid statements like 'Based on the context, ...' or 'The context information ...' or anything along those lines.

Output format:
Your output must be a single line in JSON such as:
{"reasoning": "Provide step by step reasoning for the answer.", "answer": "Provide the final answer in 2-4 words."}
"""

USER_PROMPT = """The context information below is provided as a set of entity-relation-entity triplets from knowledge graph.
---------------------
{context}
---------------------
Given the context information and not prior knowledge, answer the question.
{question}
"""

def make_qa_chat(context: str, question: str) -> list[dict]:
    return [
            {
                "role": "system",
                "content": DEFAULT_SYSTEM_PROMPT,
            },
            {
                "role": "user",
                "content": USER_PROMPT.format(context=context, question=question),
            },
        ]

# %% ../../nbs/qa.llm.ipynb 6
def parse_llm_generation(output: str):
    return json.loads(output)

# %% ../../nbs/qa.llm.ipynb 7
class QuestionAnsweringResult(BaseModel):
    """Data model for answering the question."""

    reasoning: str = Field(description="Concise reasoning for the answer.")
    answer: str = Field(description="The answer to the question in 2-4 words.")
    raw_output: str = Field(description="The raw output from the model.")


def make_question_answer_func(
    model_name: str = "gpt-3.5-turbo",
    client: OpenAI = None,
    completion_kwargs: dict | None = None,
):
    if client is None:
        client = OpenAI()

    if completion_kwargs is None:
        completion_kwargs = {}

    def func(context: str, question: str) -> QuestionAnsweringResult:
        messages = make_qa_chat(context, question)
        chat_completion = client.chat.completions.create(
            model=model_name,
            messages=messages,
            **completion_kwargs,
        )
        text = chat_completion.choices[0].message.content
        output = parse_llm_generation(text)
        return QuestionAnsweringResult(answer=output["answer"], reasoning=output["reasoning"], raw_output=text)

    return func


# %% ../../nbs/musique.multihopjerx.ipynb 6
class BaselineMultiHop:
    def __init__(self, qa_func, retrieval_func):
        self.qa_func = qa_func
        self.retrieval_func = retrieval_func

    def _call(self, example) -> QuestionAnsweringResult:
        docs = list(make_docs(example))
        
        # First question
        question1 = example["question_decomposition"][0]["question"]
        query1 = question1
        docs1 = self.retrieval_func(docs, query1)
        context1 = "\n".join(doc['text'] for doc in docs1)
        result1 = self.qa_func(context=context1, question=question1)
        hop1 = {
            "question": question1,
            "query" : query1,
            "context": context1,
            "answer": result1.answer,
            "reasoning": result1.reasoning,
        }

        # Second question
        if result1.answer == "N/A":
            return QuestionAnsweringResult({
                "answer": "N/A",
                "reasoning": result1.reasoning,
                "hops": [hop1],
            })

        question2 = example["question_decomposition"][1]["question"]
        question2 = question2.replace("#1", result1.answer)
        query2 = question2
        docs2 = self.retrieval_func(docs, query2)
        context2 = "\n".join(doc['text'] for doc in docs2)
        result2 = self.qa_func(context=context2, question=question2)
        hop2 = {
            "question": question2,
            "query": query2,
            "context": context2,
            "answer": result2.answer,
            "reasoning": result2.reasoning,
        }
        return QuestionAnsweringResult(answer=result2.answer, reasoning=result2.reasoning, raw_output=json.dumps([hop1, hop2]))

    def __call__(self, example, ignore_errors: bool = False) -> QuestionAnsweringResult:
        try:
            output = self._call(example)
        except Exception as exc:
            if ignore_errors:
                id = example['id']
                print(f"Failed to answer the question {id}\n{exc}")
                output = QuestionAnsweringResult(reasoning="", answer="N/A", raw_output=str(exc))
            else:
                raise
        return output

# %% ../../nbs/musique.multihopjerx.ipynb 7
def benchmark(
    dataf: pd.DataFrame,
    qa_func: Callable,
    retrieval_func: Callable,
    ignore_errors: bool = False,
) -> tuple[pd.DataFrame, dict]:
    pipeline = BaselineMultiHop(qa_func, retrieval_func)

    def process(example):
        output = pipeline(example, ignore_errors=ignore_errors)
        example["predicted_answer"] = output.answer
        example["raw_llm_output"] = output
        return example

    dataf = dataf.progress_apply(process, axis=1)
    dataf = compare_answers(dataf)
    scores = calculate_metrics(dataf)
    scores["fuzzy_match"] = dataf["fuzzy_match"].mean()
    return dataf, scores



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/torch.dataset.utils.ipynb.

# %% auto 0
__all__ = ['ListDataset']

# %% ../../../nbs/torch.dataset.utils.ipynb 3
from torch.utils.data import Dataset

# %% ../../../nbs/torch.dataset.utils.ipynb 4
class ListDataset(Dataset):
     def __init__(self, original_list):
        self.original_list = original_list
     
     def __len__(self):
        return len(self.original_list)

     def __getitem__(self, i):
        return self.original_list[i]



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/musique.eval.ipynb.

# %% auto 0
__all__ = ['calculate_metrics', 'compare_answers']

# %% ../../nbs/musique.eval.ipynb 3
import evaluate
import pandas as pd

from ..text.utils import fuzzy_match

# %% ../../nbs/musique.eval.ipynb 4
def calculate_metrics(dataf: pd.DataFrame) -> dict:
    metric = evaluate.load("bdsaglam/musique")
    predictions = dataf["predicted_answer"].tolist()
    references = dataf["answers"].tolist()
    scores = metric.compute(predictions=predictions, references=references)
    return scores

# %% ../../nbs/musique.eval.ipynb 5
def _exact_match(example):
    pred = example['predicted_answer']
    return pred is not None and any(pred == ref for ref in example['answers'])

def _fuzzy_match(example):
    pred = example['predicted_answer']
    return pred is not None and any((pred in ref) or (ref in pred) or fuzzy_match(pred, ref) for ref in example['answers'])


def compare_answers(dataf: pd.DataFrame) -> pd.DataFrame:
    dataf['exact_match'] = dataf.apply(_exact_match, axis=1)
    dataf['fuzzy_match'] = dataf.apply(_fuzzy_match, axis=1)
    return dataf



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/ml.experiment.ipynb.

# %% auto 0
__all__ = ['make_experiment_dir', 'before_experiment', 'after_experiment', 'make_run_experiment_sweep', 'main']

# %% ../../nbs/ml.experiment.ipynb 3
import os
from pathlib import Path
import wandb
import json
from ..utils import NestedDict, flatten_dict

# %% ../../nbs/ml.experiment.ipynb 4
def make_experiment_dir(root=".experiments", name=None):
    if name is None:
        from bellek.utils import generate_time_id
        name = generate_time_id()
    experiment_dir = Path(root) / name
    experiment_dir.mkdir(parents=True, exist_ok=True)
    return experiment_dir

# %% ../../nbs/ml.experiment.ipynb 5
def before_experiment(wandb_run):
    config = NestedDict.from_flat_dict(wandb_run.config)

    # W&B variables
    config.set("wandb.id", wandb_run.id)
    os.environ["WANDB_PROJECT"] = wandb_run.project
    os.environ["WANDB_LOG_MODEL"] = "end"

    # Set random seed
    if seed := config.get("seed"):
        from fastai.torch_core import set_seed

        set_seed(seed)

    pp_config = config.at("metaconfig.preprocessing")
    
    # Resolve paths
    if pp_config.get("resolve_paths"):
        exclude_resolving_paths = pp_config.get("exclude_resolving_paths", [])
        for k, v in config.flat().items():
            if isinstance(k, str) and k.endswith("path") and k not in exclude_resolving_paths:
                config.set(k, str(Path(v).resolve()))

    # Resolve device
    if "device" not in config and pp_config.get("resolve_device"):
        import torch

        config["device"] = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    # Update W&B config
    wandb_run.config.update(flatten_dict(config), allow_val_change=True)

    # Save preprocessed config
    with open("./config.before.json", "w") as f:
        json.dump(config, f, indent=2)


def after_experiment(wandb_run):
    config = NestedDict.from_flat_dict(wandb_run.config)
    with open("./config.after.json", "w") as f:
        json.dump(config, f, indent=2)

# %% ../../nbs/ml.experiment.ipynb 6
def make_run_experiment_sweep(run_experiment, config_defaults):
    def func():
        wandb_params = config_defaults["wandb"]
        with wandb.init(config=flatten_dict(config_defaults), **wandb_params) as wandb_run:
            before_experiment(wandb_run)
            run_experiment(wandb_run)
            after_experiment(wandb_run)
    return func

def main(run_experiment, args):
    with open(args.cfg) as f:
        config = NestedDict(json.load(f))

    is_sweep = hasattr(args, "sweep_cfg") and args.sweep_cfg
    if is_sweep:
        with open(args.sweep_cfg) as f:
            sweep_config = json.load(f)
    else:
        sweep_config = {}

    run_experiment_sweep = make_run_experiment_sweep(run_experiment, config)
    if is_sweep:
        wandb_params = config["wandb"]
        count = sweep_config.pop("count") if "count" in sweep_config else None
        sweep_id = wandb.sweep(
            sweep_config,
            entity=wandb_params["entity"],
            project=wandb_params["project"],
        )
        wandb.agent(sweep_id, run_experiment_sweep, count=count)
    else:
        run_experiment_sweep()




# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/lang.qdecomp.ipynb.

# %% auto 0
__all__ = ['QUESTION_DECOMPOSITION_SYSTEM_PROMPT_TEMPLATE', 'make_chat_prompt_template', 'parse_sub_questions',
           'make_question_decomposer']

# %% ../../nbs/lang.qdecomp.ipynb 3
from langchain_openai import ChatOpenAI
from langchain_core.output_parsers import StrOutputParser
from langchain.prompts.chat import (
    ChatPromptTemplate,
    SystemMessagePromptTemplate,
    HumanMessagePromptTemplate,
)

# %% ../../nbs/lang.qdecomp.ipynb 4
QUESTION_DECOMPOSITION_SYSTEM_PROMPT_TEMPLATE = """
Decompose the given question into 2 sub-questions such that when sub-questions are answered, the original question can be answered correctly.
The second subquestion must refer to the answer of the first question by `#1` as in the examples below. Do not create open-ended sub-questions like "Who is ..." or "How is ...".

Question: What year saw the creation of the region where the county of Hertfordshire is located?
Sub-questions:
1. In which state is Hertfordshire located?
2. When was #1 birthed?

Question: When was the institute that owned The Collegian founded?
Sub-questions:
1. Which institute does own The Collegian?
2. When #1 founded?

""".strip()


def make_chat_prompt_template() -> ChatPromptTemplate:
    return ChatPromptTemplate.from_messages(
        [
            SystemMessagePromptTemplate.from_template(QUESTION_DECOMPOSITION_SYSTEM_PROMPT_TEMPLATE),
            HumanMessagePromptTemplate.from_template("Question: {question}"),
        ]
    )


def parse_sub_questions(output: str):
    flag = False
    for line in output.splitlines():
        if line.lower().startswith("sub-questions"):
            flag = True
            continue
        if flag:
            yield line.split(".", 1)[-1].strip()


def make_question_decomposer(llm=None):
    if llm is None:
        llm = ChatOpenAI(temperature=0, model="gpt-3.5-turbo")
    
    chain = make_chat_prompt_template() | llm | StrOutputParser()
    
    def func(question):
        out = chain.invoke(dict(question=question))
        return list(parse_sub_questions(out))
    return func



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/jerx.dataset.webnlg.ipynb.

# %% auto 0
__all__ = ['batch_transform_webnlg']

# %% ../../../nbs/jerx.dataset.webnlg.ipynb 3
from ...utils import split_camel_case

# %% ../../../nbs/jerx.dataset.webnlg.ipynb 4
def _transform_relation(relation: str):
    return " ".join([word.lower() for word in split_camel_case(relation)]).strip()


def _transform_entity(entity: str):
    return entity.replace("_", " ").strip()


def _transform_triplet(triplet_string: str):
    delimiter = " | "
    triplet_string = triplet_string.replace('"', "").replace("''", "")
    entity1, relation, entity2 = triplet_string.split(delimiter)
    relation = _transform_relation(relation)
    entity1 = _transform_entity(entity1)
    entity2 = _transform_entity(entity2)
    return delimiter.join([entity1, relation, entity2])


def _batch_transform_webnlg(examples):
    for lex, mts in zip(examples["lex"], examples["modified_triple_sets"]):
        for text in lex["text"]:
            triplets = [_transform_triplet(triplet_string) for triplet_string in mts["mtriple_set"][0]]
            yield dict(text=text, triplets=triplets)


def batch_transform_webnlg(examples):
    records = list(_batch_transform_webnlg(examples))
    return {
        "text": [record["text"] for record in records],
        "triplets": [record["triplets"] for record in records],
    }



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/hf.transformers.experiment.ipynb.

# %% auto 0
__all__ = ['log', 'prepare_config_for_fp', 'preprocess_config', 'make_datacollator', 'prepare_model_for_training',
           'calculate_token_counts', 'fine_tune', 'prepare_model_for_inference', 'make_pipeline', 'predict',
           'evaluate_']

# %% ../../../nbs/hf.transformers.experiment.ipynb 3
from copy import deepcopy
from math import ceil
from typing import Any, Callable

import torch
from datasets import Dataset
from transformers import TrainingArguments, pipeline
from transformers.pipelines.text_generation import Chat
from trl import DataCollatorForCompletionOnlyLM, SFTTrainer

from .generation import generate
from .utils import load_tokenizer_model
from ..datasets.utils import load_datasets
from ...lang.dataset import partition_input_output_messages
from ...logging import get_logger
from ...utils import NestedDict, generate_time_id
from ...torch.dataset.utils import ListDataset

log = get_logger(__name__)

# %% ../../../nbs/hf.transformers.experiment.ipynb 4
def prepare_config_for_fp(config: NestedDict):
    if not torch.cuda.is_available():
        return config

    # Set float precision
    if config.at("pretrained_model.torch_dtype") in {"float16", "bfloat16"}:
        major, _ = torch.cuda.get_device_capability()
        gpu_supports_bf = major >= 8
        if gpu_supports_bf:
            log.info("GPU supports bfloat16.")
        else:
            log.info("GPU does not support bfloat16.")
        
        if config.at("pretrained_model.torch_dtype") == "bfloat16" and gpu_supports_bf:
            log.info("Using bfloat16.")
            torch_dtype, bf16, fp16, bnb_4bit_compute_dtype = ("bfloat16", True, False, "bfloat16")
        else:
            log.info("Using float16.")
            torch_dtype, bf16, fp16, bnb_4bit_compute_dtype = ("float16", False, True, "float16")
    else:
        log.info("Not using half-precision float.")
        torch_dtype, bf16, fp16, bnb_4bit_compute_dtype = (None, None, None, None)

    if config.at("pretrained_model.torch_dtype"):
        config.set("pretrained_model.torch_dtype", torch_dtype)
    if config.at("pretrained_model.quantization_config.load_in_4bit"):
        config.set("pretrained_model.quantization_config.bnb_4bit_compute_dtype", bnb_4bit_compute_dtype)
    if config.at("trainer.training_args.bf16") or config.at("trainer.training_args.fp16"):
        config.set("trainer.training_args.bf16", bf16)
        config.set("trainer.training_args.fp16", fp16)
    return config

def preprocess_config(config: NestedDict):
    config = deepcopy(config)

    if isinstance(config.at("dataset.train"), dict):
        config.set("dataset.train", [config.at("dataset.train")])
    if isinstance(config.at("dataset.validation"), dict):
        config.set("dataset.validation", [config.at("dataset.validation")])
    
    if config.at("distributed_training"):
        from accelerate import PartialState
        config.set("pretrained_model.device_map", {"": PartialState().process_index})

    config = prepare_config_for_fp(config)
    
    # Generate unique model id for output model
    if (out_model_id := config.at("hfhub.model_id")) and config.at("metaconfig.preprocessing.unique_hfhub_model_id"):
        if "-peft" not in out_model_id and config.at("trainer.lora"):
            out_model_id += "-peft"
        if wandb_run_id := config.at("wandb.id"):
            out_model_id += f"-{wandb_run_id}"
        else:
            out_model_id += f"-{generate_time_id()}"
        config.set("hfhub.model_id", out_model_id)

    return config


# %% ../../../nbs/hf.transformers.experiment.ipynb 6
def make_datacollator(tokenizer, response_template: str | None, response_template_context: str | None = None):
    if not response_template:
        return None

    if response_template_context is None:
        log.info(f"Creating completion-only data collator with response template '{response_template}'")
        data_collator = DataCollatorForCompletionOnlyLM(response_template, tokenizer=tokenizer)
    else:
        log.info(f"Creating completion-only data collator with response template '{response_template}' and context '{response_template_context}'")
        response_template_with_context = response_template_context + response_template
        response_template_ids = tokenizer.encode(response_template_with_context, add_special_tokens=False)[len(response_template_context):] 
        data_collator = DataCollatorForCompletionOnlyLM(response_template_ids, tokenizer=tokenizer)
    
    return data_collator

# %% ../../../nbs/hf.transformers.experiment.ipynb 7
def prepare_model_for_training(tokenizer, model):
    model_id = model.name_or_path.lower()

    if "llama-2" in model_id:
        from bellek.hf.transformers.llama2 import prepare_llama2_for_training

        log.info("Base model is a llama-2 model, preparing it for training.")
        prepare_llama2_for_training(tokenizer, model)
    
    elif "llama-3" in model_id:
        from bellek.hf.transformers.llama3 import prepare_llama3_for_training

        log.info("Base model is a llama-3 model, preparing it for training.")
        prepare_llama3_for_training(tokenizer, model)
    
    else:
        log.warning(f"Base model '{model_id}' is not a llama-2 or llama-3 model, no special preparation is done.")

# %% ../../../nbs/hf.transformers.experiment.ipynb 8
def calculate_token_counts(
    tokenizer,
    dataset: Dataset,
    *,
    messages_field: str = "messages",
):
    if messages_field not in dataset.column_names:
        raise ValueError(
            f"Dataset must have `{messages_field}` columns if `text_field` is not specified."
        )

    text_field = "text"
    dataset = dataset.map(
        lambda example: {
            text_field: tokenizer.apply_chat_template(
                example[messages_field],
                tokenize=False,
                add_generation_prompt=False,
            )
        }
    )

    # Inspect token counts
    tokenized_train_ds = dataset.map(lambda examples: tokenizer(examples[text_field]), batched=True)
    token_counts = [len(input_ids) for input_ids in tokenized_train_ds["input_ids"]]
    log.info(f"Input token counts: min={min(token_counts)}, max={max(token_counts)}")
    return token_counts

# %% ../../../nbs/hf.transformers.experiment.ipynb 9
def fine_tune(config: NestedDict):
    from peft import LoraConfig

    # Base model
    pretrained_model_config = config["pretrained_model"]
    model_id = pretrained_model_config.pop("model_name_or_path")
    tokenizer, base_model = load_tokenizer_model(model_id, **pretrained_model_config)
    log.info(f"Loaded base model {model_id}")

    # Prepare model for training
    prepare_model_for_training(tokenizer, base_model)

    # Train dataset
    train_ds = load_datasets(config.at("dataset.train")).shuffle(seed=config.at("seed"))
    log.info(f"Loaded training dataset with {len(train_ds)} samples.")

    # Inspect token counts
    dataset_text_field = config.at("trainer.dataset_text_field")
    token_counts = calculate_token_counts(tokenizer, train_ds)
    log.info(f"Input token counts: min={min(token_counts)}, max={max(token_counts)}")

    # Supervised fine-tuning
    if config.at("trainer.max_seq_length") is None:
        config.set("trainer.max_seq_length", ceil(max(token_counts) / 8) * 8)
    max_seq_length = config.at("trainer.max_seq_length")
    log.info(f"Setting max_seq_length={max_seq_length}")

    peft_config = LoraConfig(**config.at("trainer.lora", {}))

    packing = config.at("trainer.packing", False)

    data_collator = make_datacollator(
        tokenizer,
        config.at("trainer.response_template"),
        config.at("trainer.response_template_context"),
    )

    training_args = TrainingArguments(
        output_dir="./results",
        **config.at("trainer.training_args"),
    )

    trainer = SFTTrainer(
        tokenizer=tokenizer,
        model=base_model,
        peft_config=peft_config,
        train_dataset=train_ds,
        dataset_text_field=dataset_text_field,
        max_seq_length=max_seq_length,
        packing=packing,
        data_collator=data_collator,
        args=training_args,
    )
    log.info("Starting training...")
    trainer.train()

    # Save trained model
    log.info("Saving model...")
    final_model_id = config.at("hfhub.model_id")
    trainer.model.push_to_hub(final_model_id)
    tokenizer.push_to_hub(final_model_id)
    log.info(f"Uploaded PEFT adapters to HF Hub as {final_model_id}")

    return trainer

# %% ../../../nbs/hf.transformers.experiment.ipynb 10
def prepare_model_for_inference(tokenizer, model):
    model_id = model.name_or_path.lower()

    if "llama-2" in model_id:
        from bellek.hf.transformers.llama2 import prepare_llama2_for_inference

        log.info("Base model is a llama-2 model, preparing it for inference.")
        prepare_llama2_for_inference(tokenizer, model)
    
    elif "llama-3" in model_id:
        from bellek.hf.transformers.llama3 import prepare_llama3_for_inference

        log.info("Base model is a llama-3 model, preparing it for inference.")
        prepare_llama3_for_inference(tokenizer, model)
    
    else:
        log.warning(f"Base model '{model_id}' is not a llama-2 or llama-3 model, no special preparation is done.")

# %% ../../../nbs/hf.transformers.experiment.ipynb 11
def _load_tokenizer_model(config: NestedDict):
    model_id = config.at("hfhub.model_id")
    kwargs = deepcopy(config.get("pretrained_model", {}))
    kwargs.pop("model_name_or_path", None)
    return load_tokenizer_model(model_id, **kwargs)


def make_pipeline(config, tokenizer, model):
    prepare_model_for_inference(tokenizer, model)
    return pipeline(
        task="text-generation",
        model=model,
        tokenizer=tokenizer,
        **config.at("inference.pipeline", {}),
    )

def predict(
    config,
    *,
    tokenizer=None,
    model=None,
):
    # Load validation dataset
    dataset_config = config.at("dataset.validation")
    assert dataset_config, "Validation dataset is not provided!"
    dataset = load_datasets(dataset_config)
    assert len(dataset) > 0, "Validation dataset is empty!"

    # Ensure the dataset has input/output columns
    cols = dataset[0].keys()
    if "input" not in cols:
        if "messages" not in dataset.column_names:
            raise ValueError("Dataset must have `messages` column if `input` column are not provided.")
        dataset = dataset.map(partition_input_output_messages).remove_columns("messages")

    # Prepare text generation pipeline
    if tokenizer is None or model is None:
        tokenizer, model = _load_tokenizer_model(config)

    # Set up pipeline
    generation_params = config.at("inference.generation_params", {})
    if "max_new_tokens" not in generation_params:
        if "output" in dataset.column_names:
            token_counts = calculate_token_counts(tokenizer, dataset, messages_field = "output")
            log.info(f"Output token counts: min={min(token_counts)}, max={max(token_counts)}")
            generation_params["max_new_tokens"] = ceil(max(token_counts) / 8) * 8
        else:
            log.info("max_new_tokens is not set.")

    pipe = make_pipeline(config, tokenizer, model)

    # Generate
    # Create chats so that transformers library doesn't override our ListDataset
    chats = ListDataset([Chat(messages) for messages in dataset['input']])
    generations = generate(
        pipe,
        chats,
        **generation_params,
    )

    # Create dataframe 
    dataf = dataset.to_pandas()
    dataf["generation"] = generations
    return dataf

# %% ../../../nbs/hf.transformers.experiment.ipynb 12
def evaluate_(
    config,
    *,
    tokenizer=None,
    model=None,
    metric_kwargs: dict | None = None,
    output_parse_fn: Callable[[str], Any] | None = None,
):
    import evaluate

    output_parse_fn = output_parse_fn or (lambda x: x)

    dataf = predict(config, tokenizer=tokenizer, model=model)

    # Parse texts
    dataf["prediction"] = dataf["generation"].map(output_parse_fn)
    dataf["reference"] = dataf["output"].map(lambda x: x[0]['content']).map(output_parse_fn)

    # Compute scores
    metric = evaluate.load(config.at("evaluation.metric"))
    metric_kwargs = metric_kwargs or {}
    scores = metric.compute(
        predictions=dataf["prediction"].values,
        references=dataf["reference"].values,
        **metric_kwargs,
    )

    return scores, dataf



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/musique.nhop.ipynb.

# %% auto 0
__all__ = ['make_docs', 'Result', 'BaselineMultiHop', 'benchmark']

# %% ../../nbs/musique.nhop.ipynb 3
import json
from typing import Callable

import pandas as pd
from tqdm.auto import tqdm
from pydantic import BaseModel

from ..jerx.reward.llm import QuestionAnsweringResult
from .eval import calculate_metrics, compare_answers

tqdm.pandas()

# %% ../../nbs/musique.nhop.ipynb 4
def make_docs(example):
    ps = example["paragraphs"]
    for p in ps:
        idx = p["idx"]
        title = p["title"]
        body = p["paragraph_text"]
        is_supporting = p["is_supporting"]
        text = f"# {title}\n{body}"
        yield dict(
            text=text,
            is_supporting=is_supporting,
            parent_id=example["id"],
            idx=idx,
        )

# %% ../../nbs/musique.nhop.ipynb 5
class Result(BaseModel):
    answer: str
    hops: list[dict]

class BaselineMultiHop:
    def __init__(self, qa_func, retrieval_func):
        self.qa_func = qa_func
        self.retrieval_func = retrieval_func

    def _call(self, example) -> Result:
        docs = list(make_docs(example))
        
        # First question
        prev_answer = None
        hops = []
        for i in range(len(example["question_decomposition"])):
            question = example["question_decomposition"][i]["question"]
            if prev_answer is not None:
                question = question.replace(f"#{i}", prev_answer)
            docs = self.retrieval_func(docs, question)
            context = "\n".join(doc['text'] for doc in docs)
            result = self.qa_func(context=context, question=question)
            hop = {
                "question": question,
                "context": context,
                "answer": result.answer,
                "reasoning": result.reasoning,
            }
            hops.append(hop)
            prev_answer = result.answer
            if result.answer == "N/A":
                break

        # Second question
        return Result(
            answer=hops[-1]['answer'],
            hops=hops,
        )

    def __call__(self, example, ignore_errors: bool = False) -> Result:
        try:
            output = self._call(example)
        except Exception as exc:
            if ignore_errors:
                id = example['id']
                print(f"Failed to answer the question {id}\n{exc}")
                output = Result(answer="N/A", hops=[dict(error=str(exc))])
            else:
                raise
        return output

# %% ../../nbs/musique.nhop.ipynb 6
def benchmark(
    dataf: pd.DataFrame,
    qa_func: Callable,
    retrieval_func: Callable,
    ignore_errors: bool = False,
) -> tuple[pd.DataFrame, dict]:
    pipeline = BaselineMultiHop(qa_func, retrieval_func)

    def process(example):
        output = pipeline(example, ignore_errors=ignore_errors)
        example["predicted_answer"] = output.answer
        example["raw_output"] = output
        return example

    dataf = dataf.progress_apply(process, axis=1)
    dataf = compare_answers(dataf)
    scores = calculate_metrics(dataf)
    scores["fuzzy_match"] = dataf["fuzzy_match"].mean()
    return dataf, scores
